package apt;

import apt.city.CityMapApplet;
import apt.vrml.*;
import java.applet.*;
import java.awt.*;
import java.util.*;
import java.net.*;

/**
 *
 * This applet is the main interface to the
 * interactive apartment project.
 * 
 * It has two main screens: a "city" view
 * showing all the apartments, and an "apartment"
 * view showing an individual apartment.
 *
 */
public class ApartmentApplet extends Applet implements AppletStub, Runnable
{
    /** "City" screen showing all apartments. */
    CityMapApplet cityScreen;
    
    /** Apartment creation screen. */
    InputScreen apartmentScreen=new InputScreen();
    
    /** Layout used to flip between screens. */
    CardLayout screens=new CardLayout();
    
    /** Gateway to server routines. */
    TurbulenceServer server;
    
    /** Key to apartment that we will need to load. */
    Object currentKey;
    
    boolean initialized=false;
    
    /** Constant used in events to note that an apartment has loaded. */
    final static int APARTMENT_LOADED=777;
    
    public void init()
    {
        Display.getDefaultDisplay().init(this);
        
        server=new TurbulenceServer(urlParam("save url"),
                                     urlParam("index url"),
                                     urlParam("apartment url"),
                                     urlParam("vrml url"),
                                     urlParam("vrml base url"));
        
        apartmentScreen.setStub(this);
        apartmentScreen.setServer(server);
        cityScreen=new CityMapApplet();
        cityScreen.setStub(this);
        cityScreen.setServer(server);

        setBackground(Color.white);
        setLayout(screens);
        add(cityScreen, "city");
        add(apartmentScreen, "apartment");
        
        int w=size().width, h=size().height;
        apartmentScreen.init(w,h, false);
        cityScreen.init(w,h);
        
        initialized=true;
    }
    
    private URL urlParam(String param)
    {
        try
        {
            return new URL(getDocumentBase(), getParameter(param));
        }
        catch (Exception e)
        {
            System.out.println("Error when getting this param: "+param);
            e.printStackTrace(System.out);
            return null;
        }
    }
   
    
    /**
     *
     * Begin by showing the city view.
     *
     */
    public void start()
    {
        showCityScreen();
    }
    
    /**
     *
     * Stop by stopping animation and other threads
     * in city and apartment screens.
     *
     */
    public void stop()
    {
        apartmentScreen.stop();
        cityScreen.stop();
    }
    
    /**
     *
     * Override paint method so we can draw a progress
     * message on the screen during initialization.
     *
     */
    public void paint(Graphics g)
    {
        if (initialized) return;
        
        g.setColor(new Color(0x003399));
        Display display=Display.getDefaultDisplay();
        g.setFont(display.plainFont);
        g.drawString("Preparing city.", 
            display.instructionX, display.instructionY);
    }
    
    /**
     *
     * Show apartment screen to show the given apartment.
     * If the given apartment is null, send user to an empty
     * screen ready to be built on.
     *
     * @param apartment the apartment to view.
     *
     */
    private void showApartmentScreen(Apartment apartment)
    {
        cityScreen.stop();
        apartmentScreen.setApartment(apartment);
        apartmentScreen.setSaveable(apartment==null);
        screens.show(this, "apartment");
        apartmentScreen.start();
    }
    
    /**
     *
     * Flip to city screen.
     *
     */
    private void showCityScreen()
    {
        apartmentScreen.stop();
        cityScreen.setLoading(false);
        screens.show(this, "city");
        cityScreen.start();
    }
    
    /**
     *
     * Handle non-action events. Currently these come only
     * from the apartment screen, and can be either "Save"
     * or "Exit" signals.
     *
     */
    public boolean handleEvent(Event e)
    {
        if (e.target==this && e.id==APARTMENT_LOADED)
        {
            showApartmentScreen((Apartment)e.arg);
            return true;
        }
        if (e.target==apartmentScreen)
        {
            if (e.id==apartmentScreen.VIEW3D)
            //URL u=applet.server.make3D((RoomInfo[])e.arg, 
            //    applet.apartmentScreen.getApartment().key());
            {
                apartmentScreen.setMessage("Making 3D apartment...");
                new Launch3DThread(this,(RoomInfo[])e.arg,
                    apartmentScreen.getTour(),
                    apartmentScreen.getApartment().key()).start();
                return true;
            }
            
            
            boolean exit=e.id==apartmentScreen.EXIT;
            boolean save=e.id==apartmentScreen.APARTMENT_SAVED;
            if (exit || save)
            {
                Apartment apartment=apartmentScreen.getApartment();
                if (save && apartment!=null && !apartment.isEmpty())
                {
                    cityScreen.addApartment(apartment);
                    cityScreen.setHighlight(apartment); // Highlight new apartment in city view.
                }
                else
                    cityScreen.setHighlight(null); // Turn off any highlighting since no new apartments.
                showCityScreen();
                return true;
            }
        }
        return super.handleEvent(e);
    }
    
    void notify3DFinished()
    {
        apartmentScreen.setMessage(null);
    }

    /**
     *
     * Handle any events that tell us to switch from
     * one screen to another.
     *
     */
    public boolean action(Event e, Object arg)
    {
        System.out.println("ApartmentApplet: "+e);
        if (e.target==cityScreen)
        {
            currentKey=arg;
            new Thread(this).start();
        }
        else if (e.target==apartmentScreen)
            showCityScreen();
        return true;
    }
    
    /**
     *
     * In the run method we load in new apartment data.
     *
     */
    public void run()
    {
        deliverEvent(new Event(this, APARTMENT_LOADED, 
            server.readApartment(currentKey)));
    }
    
    /**
     *
     * Just here so we can implement the AppletStub interface.
     *
     */
    public void appletResize(int width, int height) {}

}