package apt;

import apt.vrml.*;
import apt.plan2d.*;

import java.net.*;
import java.io.*;
import java.util.*;
import java.awt.Rectangle;


/**
 *
 * The TurbulenceServer class implements the ServerGateway
 * interface for the Turbulence web site version of
 * the apartment project.
 *
 */
class TurbulenceServer implements ServerGateway
{
    /** URL of server program to save an apartment. */
    private URL saveURL;
    
    /** URL of index file. */
    private URL indexURL;
    
    /** URL of directory that holds apartment text files. */
    private URL apartmentURL;
    
    /** URL of servlet that creates VRML file for apartment. */
    private URL vrmlURL;
    
    /** String that is base for VRML files. */
    private URL vrmlBaseURL;
    
    private static final String CRLF=""+((char)13)+((char)10);
    
    private static final double DEFAULT_RATIO=500/425.0;
    
    /** Cache of loaded apartments. */
    private Hashtable apartmentCache=new Hashtable();
    
    
    /**
     *
     * Construct with the URL of the servlet to be called.
     *
     */
    public TurbulenceServer(URL saveURL, 
                            URL indexURL, 
                            URL apartmentURL,
                            URL vrmlURL,
                            URL vrmlBaseURL)
    {
        this.saveURL=saveURL;
        this.indexURL=indexURL;
        this.apartmentURL=apartmentURL;
        this.vrmlURL=vrmlURL;
        this.vrmlBaseURL=vrmlBaseURL;
    }
    
    /**
     *
     * Create a VRML file for this apartment on the server,
     * and return the URL.
     *
     */
    public URL make3D(RoomInfo[] nullInfo3D, Vector tour, String key) throws Exception
    {
        System.out.println("make3D called.");
        URL make3DURL=vrmlURL;
        
        int i;
        // make an array of room infos without null values.
        int nullCount=0;
        for (i=0; i<nullInfo3D.length; i++)
            if (nullInfo3D[i]==null) nullCount++;
            
        if (nullCount== nullInfo3D.length)
            throw new NullPointerException("make3D: got entirely null room set.");
        RoomInfo[] info3D=new RoomInfo[nullInfo3D.length-nullCount];
        int nc=0;
        for (i=0; i<nullInfo3D.length; i++)
            if (nullInfo3D[i]!=null) 
                info3D[nc++]=nullInfo3D[i];
        
        String data=RoomIO.writeAllToString(info3D);
        
        // print out infos in order.
        for (i=0; i<info3D.length; i++)
            System.out.println(i+": "+info3D[i]);
        
        String tourString="";
        Room last=null;
        for (i=0; i<tour.size(); i++)
        {
            int index=-1;
            Room r=(Room)tour.elementAt(i);
            if (r.size<1 || last==r) continue;
            for (int j=0; j<info3D.length; j++)
                if (info3D[j]!=null && info3D[j].getName().equals(r.getName()))
                    index=j;
            if (index==-1)
                continue;
            tourString+=","+index;
            last=r;
        }
        // print out tour actually being sent.
        System.out.println("tour="+tourString);

        System.out.println("TurbulenceServer.make3D: calling "+make3DURL);
        
        DataInputStream d=post(make3DURL, "id="+key+"&tour="+tourString+"&roominfo="+
            URLEncoder.encode(data)+CRLF);

        System.out.println("about to read results.");
        String line=null;
        String first=d.readLine();
        System.out.println("read >>>"+first);
        while (null != (line=d.readLine()))
            System.out.println("read >>>"+line);
            
        d.close();

        URL base1=new URL(vrmlURL, vrmlBaseURL.toString());
        URL fileURL=new URL(base1+key+".wrl");
        System.out.println("3d file:"+fileURL);
        return fileURL;
        
    }
     
     /* old: just sent word surfaces.
    public URL make3D(Hashtable info3D, String key) throws Exception
    {
        System.out.println("make3D called.");
        URL make3DURL=vrmlURL;
        apt.plan3d.WordSurface[] surfaces=WordSurfaceIO.makeArray(info3D);
        String data=WordSurfaceIO.writeAllToString(surfaces);

        System.out.println("calling "+make3DURL);
        
        DataInputStream d=post(make3DURL, "id="+key+"&surfaces="+
            URLEncoder.encode(data)+CRLF);

        System.out.println("about to read results.");
        String line=null;
        String first=d.readLine();
        System.out.println("read >>>"+first);
        while (null != (line=d.readLine()))
            System.out.println("read >>>"+line);
            
        d.close();

        URL base1=new URL(vrmlURL, vrmlBaseURL.toString());
        URL fileURL=new URL(base1+key+".wrl");
        System.out.println("3d file:"+fileURL);
        return fileURL;
        
    }
    */
    
    /**
     *
     * Read apartment summaries from server.
     *
     */
    public Vector readIndex() throws Exception
    {
        System.out.println("Reading index from:");
        System.out.println(indexURL);
        Vector v=new Vector();
        InputStream in=indexURL.openStream();
        DataInputStream data=new DataInputStream(in);
        
        Apartment apartment;
        while (null != (apartment=readNextSummary(data)))
            v.addElement(apartment);

        data.close();
        in.close();
        
        return v;
    }
    
    /**
     *
     * An apartment "summary" is all the information in an apartment
     * aside from the list of words. This method reads a single
     * summary from a DataInputStream; it's used when reading in a 
     * long list of summaries. If a problem is encountered or the
     * end of the file has been reached, the routine will return a null value.
     * <P>
     * A typical summary looks like:
     *      title	Celebrity Profile
     *      size	23
     *      seed	957076237880
     *      [window,46,201,76,68]
     *      [hall,329,60,106,96]
     *          (more rooms, etc.)
     *
     * @param data the DataInputStream to parse.
     * @return an Apartment object, or null if there was an error or EOF.
     *
     */
    private static Apartment readNextSummary(DataInputStream data)
    {
        String line="[start]";
        try
        {
            Apartment apartment=new Apartment();

            // title
            line=data.readLine();
            
            while (line!=null && line.indexOf("=")==0)
                line=data.readLine();
            if (line==null)
                return null;
            apartment.title=getValue(line, "title");

            // size
            line=data.readLine();
            apartment.size=Integer.parseInt(getValue(line, "size"));

            // seed
            line=data.readLine();
            apartment.seed=Long.parseLong(getValue(line, "seed"));
            apartment.setDate(new Date(apartment.seed));
            
            // ratio
            try
            {
                line=data.readLine();
                apartment.setRatio(Double.valueOf(getValue(line, "ratio")).doubleValue());
            }
            catch (NullPointerException e)
            {
                apartment.setRatio(DEFAULT_RATIO);
            }
            
            boolean backwardCompatibility=line.indexOf("[")==0;
            
            // rectangles
            while (true)
            {
                if (!backwardCompatibility)
                    line=data.readLine();
                else
                    backwardCompatibility=false;
                if (line==null) break;
                if (line.trim().length()==0) continue;
                char b=line.charAt(0);
                if (b=='=') break;
                if (b!='[') continue;
                StringTokenizer t=new StringTokenizer(line,"[,] ");
                String room=t.nextToken();
                int x=parseNext(t);
                int y=parseNext(t);
                int w=parseNext(t);
                int h=parseNext(t);
                apartment.addRectangle(room, new Rectangle(x,y,w,h));
            }
            apartment.calibrate();

            return apartment;
        }
        catch (Exception e)
        {
            System.out.println("problem line: "+line);
            System.out.println("readOne: "+e);
            return null;
        }
    }
    
    private static String getValue(String s, String value)
    {
        try
        {
            StringTokenizer t=new StringTokenizer(s, "\t");
            t.nextToken();
            value=t.nextToken().trim();
            return value;
        }
        catch (Exception e)
        {
            System.out.println("Couldn't get "+value+" from: "+s);
            return null;
        }
    }

    private static int parseNext(StringTokenizer t)
    {
        return Integer.parseInt(t.nextToken());
    }
    
    /**
     *
     * Load an apartment given its key.
     * Try the local cache before loading from server.
     *
     * @param key the key designating the apartment to be loaded.
     *
     */
    public Apartment readApartment(Object key)
    {
        try
        {
            if (key==null) return null;
            Object a=apartmentCache.get(key);
            if (a!=null)
                return (Apartment)a;
            URL url=new URL(apartmentURL+(key+".txt"));
            Apartment apartment=loadFromURL(url);
            apartmentCache.put(key, apartment);
            return apartment;
        }
        catch (Exception e)
        {
            System.out.println(e);
            return null;
        }
    }
    
    /**
     *
     * Save an apartment with the given title, random number seed,
     * and words.
     *
     * @param apartment the Apartment object to be saved.
     *
     */
    public void saveApartment(Apartment apartment) throws Exception
    {
        if (apartment==null || apartment.isEmpty())
            return;
            
        String title=apartment.getTitle();
        long seed=apartment.getSeed();
        Vector words=apartment.getWords();
        Vector rooms=apartment.getRoomDescriptions();
        System.out.println("calling "+saveURL);
        System.out.println("seed="+seed);
        StringBuffer data=new StringBuffer();
        data.append("title="+title+"&seed="+seed+"&size="+words.size());
        data.append("&ratio="+apartment.getRatio());
        data.append("&date="+apartment.getDate());
        
        System.out.println("date="+apartment.getDate());
        System.out.println("ratio="+apartment.getRatio());
        
        data.append("&words=");
        int i;
        for (i=0; i<words.size(); i++)
        {
            data.append((String)words.elementAt(i));
            data.append(" ");
        }
        data.append("&rooms=");
        for (i=0; i<rooms.size(); i++)
        {
            data.append((String)rooms.elementAt(i)+CRLF);
        }
        
        /*
        URLConnection conn = saveURL.openConnection();
        conn.setDoOutput(true); 
        conn.setDoInput(true); 
        PrintStream ps = new PrintStream(conn.getOutputStream());
        ps.print(post.toString());
        ps.close();
        */
        DataInputStream d=post(saveURL, data);
        displayAndClose(d);
        System.out.println("closed input stream.");
        apartmentCache.put(apartment.key(), apartment);
    }
    
    private void displayAndClose(DataInputStream d) throws Exception
    {
        System.out.println("Got this response:");
        String line=null;
        while (null != (line=d.readLine()))
            System.out.println(line);
        d.close();
        System.out.println("End response:");
    }
    
    private DataInputStream post(URL u, Object data) throws Exception
    {
        System.out.println("About to post to: "+u);
        URLConnection conn = u.openConnection();
        conn.setDoOutput(true);
        conn.setUseCaches(false);
        conn.setDefaultUseCaches(false);
        conn.setRequestProperty("Content-Type",
            "application/x-www-form-urlencoded");
        PrintStream ps = new PrintStream(conn.getOutputStream());
        ps.print(data.toString());
        ps.close();
        
        System.out.println("sent all data.");
        
        InputStream in=conn.getInputStream();
        DataInputStream d=new DataInputStream(in);
        
        System.out.println("created input stream.");
        return d;
    }
    
    /** 
     *
     * Load an Apartment from the specified URL.
     *
     * @return an Apartment constructed from the contents of the file.
     *
     */
    private static Apartment loadFromURL(URL url) throws Exception
    {
        NetTextResource info=new NetTextResource(url);
        Apartment apartment=new Apartment();
        apartment.title=info.get("title");
        apartment.seed=Long.parseLong(info.get("seed"));
        
        try
        {
            apartment.setDate(new Date(info.get("date")));
        } catch (Exception e)
        {
            System.out.println("couldn't read date.");
            apartment.setDate(new Date("9/1/2000"));
        }
        
        try
        {
            apartment.setRatio(Double.valueOf(info.get("ratio")).doubleValue());
        } catch (Exception e)
        {
            System.out.println("couldn't read ratio.");
            apartment.setRatio(DEFAULT_RATIO);
        }
        
        StringTokenizer t=new StringTokenizer(info.get("words"));
        while (t.hasMoreTokens())
            apartment.addWord(t.nextToken());
        return apartment;
    }
    
}