package apt.city;

import apt.*;
import java.applet.*;
import java.awt.*;
import java.util.*;

/** 
 *
 * The CityMapAxisPlot draws a scatterplot-style
 * map of a list of apartments. It also contains
 * code for performing an animated switch between
 * different axis measures, as well as the ability
 * to mark certain apartments as visited.
 *
 */
class CityMapAxisPlot extends BufferPanel implements Runnable
{
    /** List of apartments. */
    private Vector apartments=new Vector();
    
    /** Names of neighborhoods. */
    private String[][] neighborhood;
    
    /** Display settings. */
    private Display display=Display.getDefaultDisplay();
    
    /** Apartment to highlight. */
    private MappedApartment highlight;

    /** Thread in which do the coordinate change animation. */
    private Thread changingCoordinates;

    /** Current apartment being loaded. */
    private MappedApartment loadingApartment;

    // Mouse coordinates.
    private Point mouse=new Point(0,0);
    
    // Scaling parameters.
    private double scale=.3;
    private int aw, ah;
    
    // Controls speed of animation.
    private double step;
    
    // Current and future axes.
    private int axis1=0, axis2=1, newAxis1=axis1, newAxis2=axis2;
    
    // Initial axes.
    private String axisName1=display.axisNames[axis1];
    private String axisName2=display.axisNames[axis2];
    
    // Words for "benchmark" icons.
    private String[] modifier={"Low", "Medium", "High"};

    public synchronized void switchAxes(int newAxis1, String axisName1,
                                        int newAxis2, String axisName2)
    {
        this.newAxis1=newAxis1;
        this.newAxis2=newAxis2;
        this.axisName1=axisName1;
        this.axisName2=axisName2;
        changingCoordinates=new Thread(this);
        changingCoordinates.start();
    }
    
    public void setNeighborhood(String[][] neighborhood)
    {
        this.neighborhood=neighborhood;
    }
    
    /**
     *
     * Tell whether to display loading message.
     *
     * @param loading whether to display a loading message.
     *
     */
    public void setLoading(boolean loading)
    {
        if (!loading)
            loadingApartment=null;
        redraw();
    }
    
    /**
     *
     * Request highlighting for a particular apartment.
     *
     * @param highlight the Apartment to highlight.
     *
     */
    public void setHighlight(Apartment highlight)
    {
        if (highlight==null)
        {
            this.highlight=null;
            return;
        }
        for (int i=0; i<apartments.size(); i++)
        {
            MappedApartment a=(MappedApartment)apartments.elementAt(i);
            if (a.apartment==highlight)
            {
                this.highlight=a;
                return;
            }
        }
        System.out.println("CityMapAxisPlot.setHighlight:");
        System.out.println("    Apartment "+highlight+"  not found.");
    }

    public void run()
    {
        for (step=.05; step<=1; step+=.05)
        {
            if (changingCoordinates!=Thread.currentThread())
                return;
            redoEverything();
            try {Thread.sleep(30);} catch (Exception e) {}
        }
        axis1=newAxis1;
        axis2=newAxis2;
        redoEverything();
    }

    private synchronized void redoEverything()
    {
        createMap();
        redraw();
    }

    public void addApartment(Apartment apartment)
    {
        apartments.addElement(new MappedApartment(apartment));
    }

    synchronized void createMap()
    {
        double xMin=1e6, xMax=-1e6;
        double yMin=1e6, yMax=-1e6;
        int i;
        for (i=0; i<apartments.size(); i++)
        {
            MappedApartment a=(MappedApartment)apartments.elementAt(i);
            double y=y(a);
            double x=x(a);
            xMin=Math.min(xMin, x);
            xMax=Math.max(xMax, x);
            yMin=Math.min(yMin, y);
            yMax=Math.max(yMax, y);
        }
        
        //System.out.println("x: "+xMin+" to "+xMax);
        //System.out.println("y: "+yMin+" to "+yMax);
        //System.out.println("   ");
        
        int w=size().width, h=size().height;
        ah=(int)(scale*size().height);
        aw=(int)(scale*size().width);
        for (i=0; i<apartments.size(); i++)
        {
            MappedApartment a=(MappedApartment)apartments.elementAt(i);
            double y=y(a);
            double x=x(a);

            a.x=(int)(w*(1-scale)*(x-xMin)/(xMax-xMin))+60; // +0???
            a.y=h-(int)(h*(1-scale)*(y-yMin)/(yMax-yMin))-60; // -120??? -60??
            
            double aptScale=.4*scale*Math.sqrt(a.apartment.size/60.0);
            if (aptScale>.4) aptScale=.4;
            a.clearScaledRectangles();
            a.w=0; a.h=0;
            for (int j=0; j<a.apartment.rectangles.size(); j++)
            {
                Rectangle r=(Rectangle)a.apartment.rectangles.elementAt(j);
                Rectangle rs=new Rectangle();
                rs.x=(int)(a.x+aptScale*r.x);
                rs.y=(int)(a.y+aptScale*r.y);
                rs.width=(int)(aptScale*r.width);
                rs.height=(int)(aptScale*r.height);
                a.addScaledRectangle(rs);
                a.w=Math.max(a.w, rs.x+rs.width-a.x);
                a.h=Math.max(a.h, rs.y+rs.height-a.y);
            }

        }
    }

    private double x(MappedApartment a)
    {
        return step*a.coordinate(newAxis1)+(1-step)*a.coordinate(axis1);
    }

    private double y(MappedApartment a)
    {
        return step*a.coordinate(newAxis2)+(1-step)*a.coordinate(axis2);
    }

    public boolean handleEvent(Event e)
    {
        if (changingCoordinates!=null && changingCoordinates.isAlive())
            return false;
        return super.handleEvent(e);
    }

    synchronized public boolean mouseMove(Event e, int x, int y)
    {
        mouse.x=x;
        mouse.y=y;
        redraw();
        return true;
    }

    synchronized public boolean mouseExit(Event e, int x, int y)
    {
        return mouseMove(e, -1, -1);
    }

    synchronized public boolean mouseDown(Event e, int x, int y)
    {
        if (loadingApartment!=null)
            return true;
        mouse.x=x;
        mouse.y=y;
        MappedApartment over=null;
        for (int i=apartments.size()-1; i>=0; i--)
        {
            MappedApartment a=(MappedApartment)apartments.elementAt(i);
            if (inside(a,mouse.x, mouse.y))
            {
                over=a;
                break;
            }
        }
        if (over==null) return true;
        over.putMark();
        loadingApartment=over;
        postEvent(new Event(this, Event.ACTION_EVENT, over.apartment));
        redraw();
        return true;
    }


    synchronized public void drawOffscreen(Graphics g)
    {
        // draw backdrop
        g.setColor(Color.white);
        int w=size().width, h=size().height;
        g.fillRect(0,0,w,h);
        
        if (apartments.size()==0)
        {
            g.setColor(display.fgcolor());
            g.setFont(display.labelFont);
            g.drawString("This is an empty city.", display.instructionX, display.instructionY);
        }

        int i;

        // find which apartment we're over
        MappedApartment over=null;
        
        // determine loading state.
        boolean loading=loadingApartment!=null;
        
        int cx=0,cy=0;
        for (i=0; i<apartments.size(); i++)
        {
            MappedApartment a=(MappedApartment)apartments.elementAt(i);
            if (loadingApartment==a || 
                !loading && inside(a,mouse.x, mouse.y))
            {
                over=a;
                cx=a.x+a.w/2;
                cy=a.y+a.h/2;
                break;
            }
        }

        // Draw axis labels for this apartment.
        if (over!=null)
        {
            g.setFont(display.titleFont);
            g.setColor(display.getTitleColor());
            int value1=(int)(100*x(over));
            int value2=(int)(100*y(over));
            g.drawLine(0,cy,w,cy);
            g.drawLine(cx,0,cx,h);
            g.drawString(axisName1+" "+value1, cx+3, 20);
            String p=axisName2+" "+value2;
            int sw=display.titleFontMetrics.stringWidth(p);
            g.drawString(p, w-3-sw, cy-3);
        }

        if (!loading)
        {
            // Draw all apartments.
            g.setColor(display.planColor);
            for (i=0; i<apartments.size(); i++)
            {
                MappedApartment a=(MappedApartment)apartments.elementAt(i);
                int d=127+a.distance(mouse.x, mouse.y)/2;
                if (d>display.palestLevel) d=display.palestLevel;
                drawApartment(g,a,display.palette[d]);
            }
            
            // Draw highlighted apartment above all other non-selected items.
            if (highlight!=null)
                drawApartment(g, highlight, display.markColor);

            
            // If we're not in a transition, draw "benchmarks" showing axis positions.
            // They fall in a 3x3 grid (low/med/high x low/med/high)
            // in squares with more than average apartments.
            
            if (neighborhood!=null && axis1==newAxis1 && axis2==newAxis2)
            {
                g.setColor(display.palette[0]);
                g.setFont(display.smallFont);
                
                int m=apartments.size();
                int enoughToDisplay=1;//Math.max(1,m/9); // how many apartments in a region warrant a benchmark?
                for (int x=0; x<3; x++)
                {
                    for (int y=0; y<3; y++)
                    {
                        String place=neighborhood[2-y][x];
                        if (place==null) continue; 
                        //if (x==1 && y==1) continue; // No benchmarks in neutral middle area.
                        int n=0;
                        int qx=0,qy=0;
                        
                        // Average coords of all apartments in quadrant.
                        for (i=0; i<m; i++)
                        {
                            MappedApartment a=(MappedApartment)apartments.elementAt(i);
                            if ((3*a.x)/w==x && (3*a.y/h==y))
                            {
                                qx+=a.x; qy+=a.y; n++;
                            }
                        }
                        
                        // If there were not enough apartments, continue.
                        if (n<enoughToDisplay) continue;

                        // Otherwise, draw benchmark at the average position.
                        int u=qx/n;
                        int v=qy/n;

                        //String s1=(x==1 ? "" : modifier[x]+" "+axisName1);
                        //String s2=(y==1 ? "" : modifier[2-y]+" "+axisName2);
                        //drawBenchmark(g,s1,s2,u,v);
                        drawBenchmark(g,place,u,v);
                    }
                }
            }
            
            
        }
        

        // Draw title of selected apartment.
        if (over!=null)
        {
            boolean above=cy>h/2;
            boolean left=cx>w/2;
            
            int tx1, tx2;
            String title=over.apartment.title;
            String date=over.apartment.getDate().toString();
            if (left)
            {
                tx1=cx-3-display.hugeFontMetrics.stringWidth(title);
                tx2=cx-3-display.smallFontMetrics.stringWidth(date);
            }
            else
            {
                tx1=cx+3;
                tx2=cx+3;
            }
            
            int ty;
            if (above)
                ty=cy-6;
            else
                ty=cy-5+display.smallFontMetrics.getHeight()+
                        display.hugeFontMetrics.getHeight();
                        
            drawApartment(g,over,display.palette[0]);
            g.setFont(display.hugeFont);
            g.drawString(title,tx1,ty-display.smallFontMetrics.getHeight());
            g.setFont(display.smallFont);
            g.drawString(date,tx2, ty);
        }

        // Draw marks on all apartments.
        if (!loading)
        {
            g.setColor(display.markColor);
            for (i=0; i<apartments.size(); i++)
            {
                MappedApartment a=(MappedApartment)apartments.elementAt(i);
                a.drawMark(g);
            }
        }
        else
        {
            g.setColor(display.palette[0]);
            g.setFont(display.hugeFont);
            g.drawString("Loading apartment...", 10, 50);
        }
        
        // Sync up.
        Toolkit.getDefaultToolkit().sync();
    }
    
    

    private void drawBenchmark(Graphics g, String s1, int x, int y)
    {
        
        int r=4, d=2*r;
        x+=r+3;
        g.drawOval(x-r-1,y-r-1,d+1,d+1);
        g.fillArc(x-r,y-r,d+1,d+1,0,90);
        g.fillArc(x-r,y-r,d+1,d+1,180,90);
        if (x>size().width/2)
            g.drawString(s1,x-d-3-display.smallFontMetrics.stringWidth(s1),y+5);
        else
            g.drawString(s1,x+d+3,y+5);
    }
    
    private void drawApartment(Graphics g, MappedApartment a, Color color)
    {
        g.setColor(color);
        for (int j=0; j<a.apartment.rectangles.size(); j++)
        {
            Rectangle r=(Rectangle)a.scaledRectangles.elementAt(j);
            g.fillRect(r.x+1,r.y+1,r.width-1,r.height-1);
        }
    }

    private boolean inside(MappedApartment a, int u, int v)
    {
        for (int j=0; j<a.apartment.rectangles.size(); j++)
        {
            Rectangle r=(Rectangle)a.scaledRectangles.elementAt(j);
            if (r.inside(u,v))
                return true;
        }
        return false;
    }

}

