package apt.plan2d;

import apt.*;
import english.syntax.WordMatcher;
import english.words.*;
import java.awt.*;
import java.util.*;

/**
 *
 * "Constructor" component for an apartment.
 * Allows the user to type in words, and creates a
 * floorplan.
 *
 */
public class ApartmentConstructor extends Panel
                          implements Runnable
{

    // 2D floorplan algorithm.
    private FloorPlan floorPlanMaker=new RandomFloorPlan();

    // Display settings.
    private Display display=Display.getDefaultDisplay();
    
    // Message to be overlaid on top of apartment.
    private String message=null;
    
    // How many sentences the user has typed. Important
    // for determining what instructions to display.
    private int linesTyped=0;

    // Room info.
    private RoomDefinition roomDefinition=new RoomDefinition();
    private Room universe=new Room();
    private Room[] rooms;
    private int roomCount;
    private Vector tour;

    // Words on screen.
    private Vector typedChars;
    private Vector floatingWords=new Vector();
    private Vector roomLabels=new Vector();
    private Vector ambientWords=new Vector();
    private Vector initialWords=new Vector();
    
    // Fast data structures for complex word animation algorithms.
    private FloatingWord[][] roomWords;
    private int[] roomSize;

    // Animation stuff
    private Thread animator;

    // Offscreen buffer
    private Image bufferImage;
    private Graphics bufferScreen;

    // Typing
    private boolean editable;
    private int keyX, keyY, startX, startY, maxW;
    private String word="";
    private int letterW=display.typeFontMetrics.stringWidth("W");
    private boolean cursor=true;
    private int cursorStartX=display.instructionX-1, cursorStartY=30;
    private boolean ready=true; // whether we're ready to start typing
    private int count=0;
    private int planHeight;
    private int planWidth;

////////////////////////////////////////////////////////////////////
// Configuration and accessors.
////////////////////////////////////////////////////////////////////

    public Room[] getRooms()
    {
        return rooms;
    }
    
    public Vector getClassifiedWords()
    {
        return floatingWords;
    }
    
    /**
     *
     * Get all the data associated with the current apartment.
     *
     * @return an Apartment representing the current apartment.
     *
     */
    public Apartment getApartment()
    {
        Apartment apartment=new Apartment();

        // Basic info:
        apartment.size=floatingWords.size();
        apartment.seed=floorPlanMaker.getSeed();
        apartment.setDate(new Date());
        apartment.setRatio(planWidth/(double)planHeight);

        // Rectangles:
        int roomCount=roomLabels.size();
        for (int i=0; i<roomCount; i++)
        {
            //if (rooms[i].size<1) continue;
            FloatingWord f=(FloatingWord)roomLabels.elementAt(i);
            apartment.addRectangle(f.word, 
                new Rectangle((int)rooms[i].x, (int)rooms[i].y,
                              (int)rooms[i].w, (int)rooms[i].h));
        }
        apartment.calibrate();

        // Words:
        int i;
        for (i=0; i<floatingWords.size(); i++)
        {
            FloatingWord f=(FloatingWord)floatingWords.elementAt(i);
            apartment.addWord(f.word);
        }
        for (i=0; i<ambientWords.size(); i++)
        {
            FloatingWord f=(FloatingWord)ambientWords.elementAt(i);
            apartment.addWord(f.word);
        }
        
        // Title:
        String title;
        int n=floatingWords.size();
        if (n==0) title="empty apartment";
        else
        {
            title="";
            for (i=0; i<3 && i<n; i++)
            {
                FloatingWord f=(FloatingWord)floatingWords.elementAt(i);
                title+=f.word+" ";
            }
            title=title.trim();
        }
        apartment.setTitle(title);

        return apartment;
    }
    
    /**
     *
     * Set message, if any, to be displayed.
     *
     * @param mesage the message to be displayed.
     *
     */
    public void setMessage(String message)
    {
        this.message=message;
    }

    // Tell about changes to floorplan.
    protected void alertUpdate()
    {
        postEvent(new ApartmentEvent(this, ApartmentEvent.UPDATE));
    }

    // Tell about a cleared floorplan.
    protected void alertClear()
    {
        postEvent(new ApartmentEvent(this, ApartmentEvent.CLEAR));
    }

    public void addInitialWord(String word)
    {
        initialWords.addElement(word);
    }

    public void init(long seed, RoomDefinition roomDefinition, int w, int h)
    {
        this.roomDefinition=roomDefinition;
        this.planHeight=h;
        this.planWidth=w;
        cursorStartY=h-1-display.typeFontMetrics.getDescent();

        int i;

        floorPlanMaker=new RandomFloorPlan(seed);

        setBackground(display.palette[255]);
        typedChars=new Vector();

        bufferImage=createImage(w,h);
        bufferScreen=bufferImage.getGraphics();
        bufferScreen.setColor(display.bgcolor());
        bufferScreen.fillRect(0,0,w,h);
        bufferScreen.setFont(display.typeFont);

        requestFocus();
        initCursor();



        Vector roomNames=roomDefinition.getRoomNames();
        roomCount=roomNames.size();
        roomWords=new FloatingWord[roomCount][];
        roomSize=new int[roomCount];
        for (i=0; i<roomCount; i++)
        {
            roomWords[i]=new FloatingWord[10];
            String name=(String)roomNames.elementAt(i);
            //System.out.println(i+"  --> "+name);
            FloatingWord f=new FloatingWord(
                name,
                null,
                display.roomFont, (int)(w*Math.random()),
                (int)(h*Math.random()), i
            );
            f.width=display.roomFontMetrics.stringWidth(name);
            f.setColor(display.roomNameColor);
            roomLabels.addElement(f);
        }

        rooms=new Room[roomCount];

        for (i=0; i<roomCount; i++)
        {
            rooms[i]=new Room(.0001);
            rooms[i].setName(display.roomNames[i]);
        }

        addInitialWords(initialWords);
        alertUpdate();
    }
    
    private void addInitialWords(Vector initialWords)
    {
        for (int i=0; i<initialWords.size(); i++)
        {
            String wd=(String)initialWords.elementAt(i);
            addWordWithoutRetiling(wd,rand(planWidth), 
                Math.random()<.5 ? -20 : planHeight+20);
            ready=true;
        }
        retile();
    }
    
    synchronized public void setApartment(Apartment apartment)
    {
        clear();
        editable=apartment==null;
        linesTyped=editable ? 0 : 10000;
        if (apartment==null)
            return;

        floorPlanMaker=new RandomFloorPlan(apartment.getSeed());
        addInitialWords(apartment.getWords());
        alertUpdate();
    }
    
    synchronized public Vector getTour()
    {
        return tour;
    }

    // Rearrange 2D plan
    synchronized void retile()
    {
        int w=getPlanWidth(), h=getPlanHeight();
        int n=rooms.length;
        Room[] array=new Room[n];
        System.arraycopy(rooms,0,array,0,n);
        tour=floorPlanMaker.tile(array,0,0,w,h-30);
/*
        // tour test
        for (int j=0; j<tour.size(); j++)
        {
            Room r=(Room)tour.elementAt(j);
            if (r.size>=1)
                System.out.println(r);
        }
*/
        universe.setRect(20,60,w-40,h-80);
        double sum=sum(rooms);
        int roomCount=roomLabels.size();
        for (int i=0; i<roomCount; i++)
        {
            FloatingWord f=(FloatingWord)roomLabels.elementAt(i);
            f.halt();
            f.setPosition(rooms[i].x+rooms[i].w/2,
                    rooms[i].y+rooms[i].h/2);
            double p=rooms[i].size/sum;
            rooms[i].wallWidth=(int)(1+12*p);
            int c=(int)(500*p);
            if (c>100) c=100;
            if (c<61) c=61;
            Color color=display.palette[255-c];
            rooms[i].color=color;
            f.setColor(color);

        }
    }

    static double sum(Room[] s)
    {
        double sum=0;
        for (int i=0; i<s.length; i++)
            sum+=s[i].size;
        return sum;
    }
    

    int getPlanHeight()
    {
        return planHeight;
    }

    int getPlanWidth()
    {
        return planWidth;
    }

    int rand(int n)
    {
        return (int)(n*Math.random());
    }

    public synchronized void clear()
    {
        message=null;
        linesTyped=0;
        for (int i=0; i<roomCount; i++)
        {
            rooms[i].size=.1;
            roomSize[i]=0;
        }
        floatingWords=new Vector();
        ambientWords=new Vector();
        typedChars=new Vector();
        floorPlanMaker=new RandomFloorPlan();
        initCursor();
        alertClear();
    }

    void initCursor()
    {
        startTypingAt(cursorStartX,cursorStartY);
    }

    synchronized void addWord(String word, int x, int y)
    {
        addWordWithoutRetiling(word,x,y);
        retile();
    }

    synchronized void addWordWithoutRetiling(String word, int x, int y)
    {
        word=word.trim();
        if (word.length()==0 || (word.length()==1 &&
            !word.equalsIgnoreCase("i")) ||
            WordReference.isConnector(word))
            return;

        int type=wordBin(word);
        FloatingWord f=new FloatingWord(word, roomDefinition.root(word),
            display.typeFont, x, y, type);
        //System.out.println("root="+f.getWord());
        f.width=display.typeFontMetrics.stringWidth(word);
        if (type>=0)
        {
            floatingWords.addElement(f);
            int s=(int)Math.round(++rooms[type].size);
            f.setColor(display.recognizedWordColor);
            f.setEnglishWord(
                roomDefinition.getWord(word).variant(word)
            );
            int n=roomWords[type].length;
            if (s>n)
            {
                FloatingWord[] bigger=new FloatingWord[2*n];
                System.arraycopy(roomWords[type],0,bigger,0,n);
                roomWords[type]=bigger;
            }
            roomWords[type][s-1]=f;
            roomSize[type]++;
        }
        else
        {
            f.setColor(display.ambientColor);
            ambientWords.addElement(f);
        }
    }

    /**
     *
     * Which room does this word belong to?
     *
     */
    private int wordBin(String word)
    {
        return roomDefinition.getRoomNumber(word);
    }

    private void startTypingAt(int x, int y)
    {
        keyX=x; keyY=y; startX=x; startY=y;
    }

//////////////////////////////////////////////////////////////////////////
// Event handling
//////////////////////////////////////////////////////////////////////////

    public synchronized boolean keyDown(Event e, int key)
    {
        int n=typedChars.size();
        if (key==8) // backspace.
        {
            
            if (n>0)
            {
                typedChars.removeElementAt(n-1);
                keyX-=letterW;
            }
            return true;
        }
        if (key==13 || key==10)
        {
            carriageReturn();
            return true;
        }
        char c=(char)e.key;
        CharPoint cp=new CharPoint(keyX, keyY, c);
        keyX+=letterW;
        maxW=Math.max(keyX-startX, maxW);
        typedChars.addElement(cp);

        if (c=='?' || c=='!' || keyX>getPlanWidth()-20 && key==' ')
        {
            carriageReturn();
        }
        
        if (c=='.')
        {
            String m=wordSoFar().toLowerCase();
            if (!(m.endsWith("mr.") || m.endsWith("ms.") ||
                  m.endsWith("ms.") || m.endsWith("dr.")))
                  carriageReturn();
        }
        
        return true;
    }

    String wordSoFar()
    {
        int n=typedChars.size();
        StringBuffer b=new StringBuffer(n);
        for (int i=0; i<n; i++)
            b.append(((CharPoint)typedChars.elementAt(i)).c);
        return b.toString();
    }

    synchronized void carriageReturn()
    {
        int wx=0, wy=0;
        String s="";
        int n=typedChars.size();
        int i;
        if (n==0) return;
        
        // Record that the user has typed another lines.
        linesTyped++;

        // break sentence
        Vector words=SentenceBreaker.divide(wordSoFar());

        // go through words, track x-coord & char #
        int startChar=0;
        for (i=0; i<words.size(); i++)
        {
            CharPoint cp=(CharPoint)typedChars.elementAt(startChar);
            String w=(String)words.elementAt(i);
            addWord(w, cp.x, cp.y);
            startChar+=w.length();
        }
        alertUpdate();
        typedChars=new Vector();
        initCursor();
    }

    String append(String s, String t)
    {
        return s.length()==0 ? t : s+" "+t;
    }

///////////////////////////////////////////////////////////////////
// Motion code
///////////////////////////////////////////////////////////////////

    
    private synchronized void updateWordPositions()
    {
        int i, n=ambientWords.size();
        long time=System.currentTimeMillis();
        for (i=0; i<n; i++)
        {
            FloatingWord word=(FloatingWord)ambientWords.elementAt(i);
            Motion.bounceWord(word, universe);
            word.move();
            word.makeOlder();
        }
            
        for (i=0; i<roomCount; i++)
        {
            n=roomSize[i];
            Room r=rooms[i];
            for (int j=0; j<n; j++)
            {
                FloatingWord word=roomWords[i][j];
                switch (i)
                {
                    case 0: // window
                        Motion.perspective(word,r,j,n,time); break;
                    case 1: // hall
                        Motion.vertical(word,r,j,n,time); break;
                    case 2: // bathroom
                        Motion.jumble(word, r,j,n); break;
                    case 3: // office
                        Motion.grid(word,r,j,n); break;
                    case 5: // library
                        Motion.align(word, r,j,n,time); break;
                    case 6: // media
                        Motion.burst(word,r,j,n,time); break;
                    case 7: // dining room
                        Motion.circle(word,r,j,n,time); break;
                    case 8: // foyer
                        Motion.implode(word,r,j,n,time); break;
                    case 9: // kitchen
                        Motion.boil(word,r,j,n,time); break;
                    case 10: // bedroom
                        Motion.sinuous(word,r,j,n,2,2.732,time); break;
                    case 11: // closet
                        Motion.staticGrid(word,r,j,n); break;
                    default: // living room
                        Motion.bounceWord(word, r);
                }
                word.move();
                word.makeOlder();
                
            }
        }
    }
    

/////////////////////////////////////////////////////////////////////////
// Drawing code
/////////////////////////////////////////////////////////////////////////


    synchronized void drawChars()
    {
        bufferScreen.setColor(display.fgcolor());
        for (int i=0; i<typedChars.size(); i++)
        {
            CharPoint cp=(CharPoint)typedChars.elementAt(i);
            bufferScreen.drawString(String.valueOf(cp.c),cp.x,cp.y);
        }
    }

    void drawOneRoom(Graphics g, Room r, boolean empty, boolean erase)
    {
        int x=(int)r.x;
        int y=(int)r.y;
        int w=(int)(r.w+r.x)-x-1;
        int h=(int)(r.h+r.y)-y-1;

        if (erase)
        {
            int s=255-(int)(10*Math.sqrt(r.size));
            if (s<150) s=150;
            g.setColor(empty ? display.emptyColor : display.planColor);
            //g.setColor(display.palette[s]);  // this shades rooms according to size.
            int j=r.wallWidth;
            g.fillRect(x+j,y+j,w-2-2*j+1,h-2-2*j+1);
            return;
        }

        g.setColor(r.color);
        for (int j=0; j<r.wallWidth; j++)
        {
            g.drawRect(x+j,y+j,w-2-2*j,h-2-2*j);
        }
    }


    void drawOneWord(Graphics g, FloatingWord word)
    {
        int n=(int)(2*Math.sqrt(word.age));
        if (n>240) n=240;
        word.draw(g);
    }

    void drawWords(Graphics g, Vector words, boolean motion)
    {
        int n=words.size();
        int i;
        
        boolean even;
        do
        {
            even=true;
            for (i=0; i<n; i++)
            {
                FloatingWord f=(FloatingWord)words.elementAt(i);
                if (f.closeToTarget())
                {
                    even=false;
                    f.snapToTarget();
                }
            }
        } while(!even);

        for (i=0; i<n; i++)
            drawOneWord(g, (FloatingWord)words.elementAt(i));

        double randCutoff=Math.max(.02, 3.0/n);

        for (i=1; i<n; i++)
        {
            FloatingWord f1=(FloatingWord)words.elementAt(i);
            if (f1.type<0 && Math.random()>randCutoff) continue;
            for (int j=0; j<i; j++)
            {
                FloatingWord f2=(FloatingWord)words.elementAt(j);
                if (f1.type!=f2.type) continue;
                if (f1.age<30 || f2.age<30) continue;
                if (f1.distance(f2)<200)
                {
                    if (null != WordMatcher.canPair(f1.englishWord,
                                        f2.englishWord))
                    {
                        if (f1.track(f2)) return;
                    }
                    else
                    if (null != WordMatcher.canPair(f2.englishWord,
                                        f1.englishWord))
                    {
                        if (f2.track(f1)) return;
                    }
                }
            }
        }
    }
    
    /**
     *
     * Erase the offscreen buffer, eliminating any previously
     * drawn versions of this object.
     *
     */
    public synchronized void erase()
    {
        bufferScreen.setColor(display.bgcolor());
        bufferScreen.fillRect(0,0,planWidth,planHeight);
    }

    protected synchronized void drawOffscreen()
    {
        erase();
        Graphics g=bufferScreen;

        drawWords(g, ambientWords, true);
        int n=roomLabels.size();
        int i;
        for (i=0; i<n; i++)
        {
            if (rooms[i].size>=1)
                drawOneRoom(g, rooms[i], false, false);
        }
        for (i=n-1; i>=0; i--)
        {
            if (rooms[i].size>=1)
                drawOneRoom(g, rooms[i], i==display.WINDOW, true); 
        }
        for (i=0; i<n; i++)
        {
            if (rooms[i].size>=1)
                drawOneWord(g, (FloatingWord)roomLabels.elementAt(i));
        }

        drawWords(g, floatingWords, true);

        g.setFont(display.typeFont);
        g.setColor(display.fgcolor());
        int s=letterW;
        int cx=cursorStartX-s, cy=cursorStartY-1;
        g.drawLine(cx,cy,cx+s/2,cy-s/2);
        g.drawLine(cx,cy-s,cx+s/2,cy-s/2);
        drawChars();

        if (System.currentTimeMillis()%500<250)
        {
            g.setColor(display.fgcolor());
            int letterH=display.typeFontMetrics.getAscent();
            g.fillRect(keyX,keyY-letterH,letterW,letterH+2);
        }
        
        if (message!=null)
        {
            Font big=display.hugeFont;
            FontMetrics bigMeasure=getFontMetrics(big);
            g.setFont(big);
            //String message="Saving apartment as \""+getApartment().getTitle()+"\"";
            int sw=bigMeasure.stringWidth(message);
            int sh=bigMeasure.getAscent();
            g.drawString(message, (planWidth-sw)/2, (planHeight+sh)/2);
        }
        
        String instructions=display.instructions(linesTyped);
        g.setFont(display.labelFont);
        g.setColor(display.fgcolor());
        g.drawString(instructions, display.instructionX, display.instructionY);
        
        Toolkit.getDefaultToolkit().sync();
    }

    public void update(Graphics g)
    {
        paint(g);
    }

    public synchronized void paint(Graphics g)
    {
        if (ready)
        {
            if (bufferImage!=null)
                g.drawImage(bufferImage,0,0,null);
        }
        else // draw message saying user should click to begin.
        {
            g.setColor(display.palette[0]);
            int w=getPlanWidth(), h=getPlanHeight();
            g.drawRect(0,0,w-1,h-1);
            g.setFont(display.typeFont);
            String intro="House of Words, 4.0. Click to begin.";
            g.drawString(intro, cursorStartX, cursorStartY);
        }

    }

///////////////////////////////////////////////////////////
// Animation threading code
///////////////////////////////////////////////////////////

    public synchronized void start()
    {
        animator=new Thread(this);
        animator.start();
    }

    public synchronized void stop()
    {
        animator=null;
    }

    public void run()
    {
        try
        {
            while (animator==Thread.currentThread())
            {
                if (ready)
                {
                    updateWordPositions();
                    drawOffscreen();
                    repaint();
                    Thread.sleep(40);
                }
                else
                    Thread.sleep(20);
            }
        }
        catch (Exception e)
        {
            System.out.println("Animation prob.");
            e.printStackTrace();
        }
    }

}

